#include <stdio.h>
#include <stdlib.h>

#include "config.h"

#ifndef NLL_OUTPUT_NOWAIT
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/select.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "nlllib.h"

int nll_istty = 0;
FILE *nll_stdin  = NULL;
FILE *nll_stdout = NULL;

void nll_error_print(const char *filename, int line,
		     int errcode, const char *param)
{
  if (!param)
    param = "(?)";

  switch (errcode) {
  case NLL_ERRCODE_FILE_NOT_FOUND:
    fprintf(nll_stdout, "File not found: %s\n", param);
    break;
  case NLL_ERRCODE_LINE_BUFFER_OVER:
    fprintf(nll_stdout, "Line buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_LINE_NOT_EMPTY:
    fprintf(nll_stdout, "Line not empty: %s\n", param);
    break;
  case NLL_ERRCODE_LINE_TOO_LONG:
    fprintf(nll_stdout, "Too long line: %s\n", param);
    break;
  case NLL_ERRCODE_LINE_NOT_FOUND:
    fprintf(nll_stdout, "Line not found: %s\n", param);
    break;
  case NLL_ERRCODE_VARIABLE_NOT_EMPTY:
    fprintf(nll_stdout, "Variable not empty: %s\n", param);
    break;
  case NLL_ERRCODE_VARIABLE_INVALID_NAME:
    fprintf(nll_stdout, "Invalid variable name: %s\n", param);
    break;
  case NLL_ERRCODE_VARIABLE_TOO_LONG:
    fprintf(nll_stdout, "Too long variable: %s\n", param);
    break;
  case NLL_ERRCODE_VARIABLE_BUFFER_OVER:
    fprintf(nll_stdout, "Variable buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_VARIABLE_NOT_FOUND:
    fprintf(nll_stdout, "Variable not found: %s\n", param);
    break;
  case NLL_ERRCODE_VALUE_NOT_EMPTY:
    fprintf(nll_stdout, "Value not empty: %s\n", param);
    break;
  case NLL_ERRCODE_VALUE_BUFFER_OVER:
    fprintf(nll_stdout, "Value buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_VALUE_INVALID_TYPE:
    fprintf(nll_stdout, "Invalid value type: %s\n", param);
    break;
  case NLL_ERRCODE_VALUE_CONST_VALUE:
    fprintf(nll_stdout, "Constant value: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_NOT_EMPTY:
    fprintf(nll_stdout, "Formula not empty: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_BUFFER_OVER:
    fprintf(nll_stdout, "Calculation buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_LESS_PARAMETER:
    fprintf(nll_stdout, "Less parameter in operand: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_MUCH_PARAMETER:
    fprintf(nll_stdout, "Much parameter in operand: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_DIV_ZERO:
    fprintf(nll_stdout, "Division by zero: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_INVALID_OPERATOR:
    fprintf(nll_stdout, "Invalid operator: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_INVALID_PARAMETER:
    fprintf(nll_stdout, "Invalid parameter: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_UNSUPPORTED_OPERATOR:
    fprintf(nll_stdout, "Unsupported operator: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_ARRAY_OVER:
    fprintf(nll_stdout, "Array size overflow: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_UNKNOWN_FUNCTION:
    fprintf(nll_stdout, "Unknown function: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_NULL_POINTER:
    fprintf(nll_stdout, "Null pointer access: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_INVALID_POINTER:
    fprintf(nll_stdout, "Invalid pointer: %s\n", param);
    break;
  case NLL_ERRCODE_FORMULA_UNSUPPORTED_FUNCTION:
    fprintf(nll_stdout, "Unsupported function: %s\n", param);
    break;
  case NLL_ERRCODE_LABEL_NOT_EMPTY:
    fprintf(nll_stdout, "Label not empty: %s\n", param);
    break;
  case NLL_ERRCODE_LABEL_BUFFER_OVER:
    fprintf(nll_stdout, "Label buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_LABEL_INVALID_NAME:
    fprintf(nll_stdout, "Invalid label name: %s\n", param);
    break;
  case NLL_ERRCODE_LABEL_TOO_LONG:
    fprintf(nll_stdout, "Too long label: %s\n", param);
    break;
  case NLL_ERRCODE_LABEL_NOT_FOUND:
    fprintf(nll_stdout, "Label not found: %s\n", param);
    break;
  case NLL_ERRCODE_LABEL_DUPLICATE:
    fprintf(nll_stdout, "Duplicate label: %s\n", param);
    break;
  case NLL_ERRCODE_POSITION_STACK_OVER:
    fprintf(nll_stdout, "Position stack overflow: %s\n", param);
    break;
  case NLL_ERRCODE_POSITION_NOT_FOUND:
    fprintf(nll_stdout, "Position not found: %s\n", param);
    break;
  case NLL_ERRCODE_POSITION_DIFFERENT_TYPE:
    fprintf(nll_stdout, "Different position type: %s\n", param);
    break;
  case NLL_ERRCODE_POSITION_DIFFERENT_VARIABLE:
    fprintf(nll_stdout, "Different position variable: %s\n", param);
    break;
  case NLL_ERRCODE_COMMAND_NOT_EMPTY:
    fprintf(nll_stdout, "Command not empty: %s\n", param);
    break;
  case NLL_ERRCODE_COMMAND_BUFFER_OVER:
    fprintf(nll_stdout, "Command buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_COMMAND_TOO_LONG:
    fprintf(nll_stdout, "Too long command: %s\n", param);
    break;
  case NLL_ERRCODE_COMMAND_NOT_FOUND:
    fprintf(nll_stdout, "Command not found: %s\n", param);
    break;
  case NLL_ERRCODE_STRING_NOT_EMPTY:
    fprintf(nll_stdout, "String not empty: %s\n", param);
    break;
  case NLL_ERRCODE_STRING_TOO_LONG:
    fprintf(nll_stdout, "Too long string: %s\n", param);
    break;
  case NLL_ERRCODE_STRING_BUFFER_OVER:
    fprintf(nll_stdout, "String buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_ARRAY_NOT_EMPTY:
    fprintf(nll_stdout, "Array not empty: %s\n", param);
    break;
  case NLL_ERRCODE_ARRAY_BUFFER_OVER:
    fprintf(nll_stdout, "Array buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_AREA_NOT_EMPTY:
    fprintf(nll_stdout, "Area not empty: %s\n", param);
    break;
  case NLL_ERRCODE_AREA_BUFFER_OVER:
    fprintf(nll_stdout, "Area buffer overflow: %s\n", param);
    break;
  case NLL_ERRCODE_AREA_OUT_OF_RANGE:
    fprintf(nll_stdout, "Out of area range: %s\n", param);
    break;
  case NLL_ERRCODE_STACK_NOT_EMPTY:
    fprintf(nll_stdout, "Stack not empty: %s\n", param);
    break;
  case NLL_ERRCODE_STACK_STACK_OVER:
    fprintf(nll_stdout, "Stack overflow: %s\n", param);
    break;
  case NLL_ERRCODE_STACK_NOT_FOUND:
    fprintf(nll_stdout, "Stack not found: %s\n", param);
    break;
  case NLL_ERRCODE_NLL_SYNTAX_ERROR:
    fprintf(nll_stdout, "Syntax error: %s\n", param);
    break;
  case NLL_ERRCODE_NLL_WORD_TOO_LONG:
    fprintf(nll_stdout, "Too long word: %s\n", param);
    break;
  case NLL_ERRCODE_NLL_NO_CONTINUE_POINT:
    fprintf(nll_stdout, "No continue point: %s\n", param);
    break;
  case NLL_ERRCODE_NLL_INVALID_COMMAND:
    fprintf(nll_stdout, "Invalid command: %s\n", param);
    break;

  case NLL_ERRCODE_UNKNOWN:
  default:
    fprintf(nll_stdout, "Unknown error: (%s:%d): %s\n", filename, line, param);
    break;
  }
}

void nll_error_exit(const char *filename, int line,
		    int errcode, const char *param)
{
  nll_error_print(filename, line, errcode, param);
  exit(1);
}

static int finished = 0;
static int finished_lock = 0;

void nll_finished_clear(void)
{
  finished = 0;
}

void nll_finish(void)
{
  finished++;
}

int nll_is_finished(void)
{
  return finished_lock ? (finished > 4) : (finished > 0);
}

int nll_finish_lock(int locked)
{
  int old;
  old = finished_lock;
  finished_lock = locked;
  return old;
}

int nll_finish_is_locked(void)
{
  return finished_lock;
}

static FILE *nll_fp = NULL;

int nll_file_open(const char *filename)
{
  FILE *fp;

  fp = fopen(filename, "r");
  if (!fp)
    return NLL_ERRCODE_FILE_NOT_FOUND;

  nll_file_close();
  nll_fp = fp;

  return 0;
}

int nll_file_close(void)
{
  if (nll_fp)
    fclose(nll_fp);
  nll_fp = NULL;
  return 0;
}

FILE *nll_file_fp(void)
{
  return nll_fp;
}

static FILE *_nll_stdin  = NULL;
static FILE *_nll_stdout = NULL;

int nll_stdin_save(FILE *fp)
{
  if (!_nll_stdin) {
    _nll_stdin = nll_stdin;
  } else {
    fclose(nll_stdin);
  }
  nll_stdin = fp;
  return 0;
}

int nll_stdin_restore(void)
{
  if (!_nll_stdin) {
    ;
  } else {
    fclose(nll_stdin);
    nll_stdin = _nll_stdin;
    _nll_stdin = NULL;
  }
  return 0;
}

int nll_stdout_save(FILE *fp)
{
  fflush(nll_stdout);
  if (!_nll_stdout) {
    _nll_stdout = nll_stdout;
  } else {
    fclose(nll_stdout);
  }
  nll_stdout = fp;
  return 0;
}

int nll_stdout_restore(void)
{
  if (!_nll_stdout) {
    ;
  } else {
    fflush(nll_stdout);
    fclose(nll_stdout);
    nll_stdout = _nll_stdout;
    _nll_stdout = NULL;
  }
  return 0;
}

int nll_wait_output(FILE *fp)
{
#ifndef NLL_OUTPUT_NOWAIT
  int r, fd;
  fd_set fds;
  struct timeval t;

  if (!fp)
    fp = nll_stdout;

  while (1) {
    fd = fileno(fp);
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    t.tv_sec = 0;
    t.tv_usec = 100000;

    r = select(fd + 1, NULL, &fds, NULL, &t);

    if (r > 0) {
      if (FD_ISSET(fd, &fds))
	break;
      continue;
    }

    if (r == 0)
      break;
  }
#endif

  return 0;
}
